classdef problem < handle
    properties
        name string

        % Continuous problem
        d uint8
        B (:, 2) double
        Ih
        oblst
        g

        % Discretization
        n int64
        N
        x
        I
        O

        % h - Grid
        h double {mustBePositive}
        sz (1, :) int64

        % Discrete cognitive maps
        T
        VO

        % Useful to store for plotting
        Tmax double {mustBePositive}
        one_dim_grids (:, 1) cell
        initial_p
    end

    methods
        function p = problem(pc, h)
            arguments
                pc problem_c
                h
            end
            p.name = pc.name;

            p.d = length(pc.B);
            p.B = pc.B;
            p.Ih = pc.Ih;
            p.oblst = pc.oblst;
            p.g = pc.g;
            
            p.h = h;
            p.sz = arrayfun(@(i) ceil((p.B(i,2)-p.B(i,1))/p.h) + 1, 1:p.d);
            p.n = prod(p.sz);
            p.N = 1:p.n;
            p.x = @(i) p.B(:,1) + p.h * double(p.s(i) - 1);

            [~, p.I] = min(arrayfun(@(i) vecnorm(p.Ih - p.x(i)), p.N));

            if numel(p.g(p.x([1 2]), 0)) == 1 %#ok<ISCL> % Check if g is defined to be vectorized
                p.O = @(inds, t) arrayfun(@(i) p.g(p.x(i), t) < 0, inds);
            else
                p.O = @(i, t) p.g(p.x(i), t) < 0; % More efficient plotting
            end

            p.Tmax = pc.Tmax;
            p.one_dim_grids = arrayfun(@(i) linspace(p.B(i, 1), p.B(i, 2), p.sz(i)), (1:p.d)', 'UniformOutput', false);
            p.initial_p = p.x(p.I);
        end


        function subscripts = s(p, i)
            arguments
                p 
                i int64
            end
            s2 = idivide(i, p.sz(1), "ceil");
            subscripts = [i-p.sz(1)*(s2-1); s2];
        end

        function i = sinv(p, subscripts)
            temp = num2cell(subscripts);
            i = zeros(1, size(temp, 2));
            for j = 1:length(i)
                try
                    i(j) = sub2ind(p.sz, temp{:, j});
                catch
                    i(j) = p.n + 1;
                end
            end
        end


        function solve(p, options)
            arguments
                p

                options.solver = @Algorithm_builder;

                options.view_process logical = 0
                options.position = [50 50 1600 900]
                options.obs_transparency double = .7

                options.FrameRate = 4
                options.duration = 10
            end

            Algorithm = options.solver(p);

            T_hat = zeros(1, p.n + 1) + Inf; T_hat(p.I) = 0;
            S = zeros(1, p.n); S(p.I) = 1; S(p.n + 1) = 2;
            
            if options.view_process                
                num_frames = ceil(options.FrameRate * options.duration);
                times_shown = [linspace(0, p.Tmax, num_frames), Inf];
                f = figure(2);
                if ~isempty(options.position)
                    f.Position = options.position;
                end
            else
                times_shown = Inf;
            end

            stop = ~ismember(1, S);
            while ~stop 
                [T_hat, S, t, stop] = Algorithm(T_hat, S, times_shown(1));
                times_shown(1) = [];
                if options.view_process
                    p.view_map(map=T_hat(1:end-1), obs_t=t, obs_transparency=options.obs_transparency, message="t = " + sprintf('%3.1f', t))
                    pause(.1)
                end
            end
        
            p.T = reshape(T_hat(1:end-1), p.sz);
            p.VO = p.calculate_VO(p.T, 1);
        end

        function VO = calculate_VO(p, map, forward)
            Discontinuities = get_discontinuity_map(map, p.h, 1.1, forward);
            Infinites = map == Inf;

            VO = or(Discontinuities, Infinites);
        end

        function view_map(p, options)
            arguments
                p
                options.map = p.T
                options.message = ""
                options.num_levels int8 = 50
                options.obs_transparency = 0
                options.obs_t
            end

            p.cntf(map=options.map, num_levels=options.num_levels)

            if options.obs_transparency
                p.view_frame(options.obs_t, transparency=options.obs_transparency)
            end

            title(options.message)
        end

        function view_map_with_discontinuities(p)
            map = p.T;
            map(p.VO) = Inf;
            map(p.calculate_VO(p.T, -1)) = Inf;
            p.view_map(map=map)
        end

        function view_frame(p, t, options)
            arguments
                p
                t
                options.transparency double = 1 
            end

            hold on
            cmap = colormap;
            cmap(end,:) = [.16 .72 .92];
            colormap(cmap)

            p.imsc(p.O(p.N, t), transparency=options.transparency);
            plot(p.initial_p(1, :), p.initial_p(2, :), 'c.', MarkerSize=20)
            hold off
        end

        function view_times(p, map, options)
            arguments
                p
                map
                options.message = ""
            end

            p.imsc(map, message=options.message)
            title(options.message)
        end

        function imsc(p, map, options)
            arguments
                p
                map
                options.transparency = "No"
                options.message = ""
            end

            map = reshape(map, p.sz)';
            map = map * p.Tmax * .25;

            if isa(options.transparency, 'string')
                imagesc(p.B(1, :), p.B(2, :), map);
            else
                imagesc(p.B(1, :), p.B(2, :), map, 'AlphaData', options.transparency*(map ~= 0));
            end

            set(gca,'YDir','normal')
            axis equal
            colorbar
            xlim(p.B(1, :))
            ylim(p.B(2, :))
            
            title(options.message, 'Interpreter','latex')
        end

        function cntf(p, options)
            arguments
                p
                options.map = p.T
                options.num_levels = 40
                options.message = ""
                options.Tmax = p.Tmax
            end
            map = options.map;
            map = reshape(map, p.sz)';
            map(map==Inf) = -1;
            levels = linspace(0, options.Tmax, options.num_levels);

            contourf(p.one_dim_grids{1}, p.one_dim_grids{2}, map, levels);
            colormap('turbo')
            clim([0, options.Tmax])
            colorbar
            xlim(p.B(1, :))
            ylim(p.B(2, :))
            axis equal
            title(options.message, 'Interpreter','latex')
        end
    end
end
